"""some additional helpers for trackunit"""

from datetime import datetime, timedelta
from copy import deepcopy
from  dateutil.parser.isoparser import isoparse

def get_datetime(time_string):
    """transforms trackunit time string to datetime"""
    #return datetime.strptime(time_string.split('.')[0],"%Y-%m-%dT%H:%M:%S")
    return isoparse(time_string)

def get_next_section(data,finsec,fendsec=None,min_insec_len=None,min_endsec_len=0):
    """
    returns the next section wich fulfills finsec with at least min_insec_len datapoints
    and ends with data wich fulfills fendsec which is at least min_endsec_len long
    """
    data.sort(key=lambda x: datetime.strptime(x['time'].split('.')[0],"%Y-%m-%dT%H:%M:%S"))
    out = []
    in_section = False
    off_sec_cnt = 0
    for datapoint in data:
        if not in_section and finsec(datapoint):
            out = []
            off_sec_cnt = 0
            out.append(datapoint)
            in_section = True
        elif in_section and finsec(datapoint):
            out.append(datapoint)
        elif in_section and not finsec(datapoint):
            if (min_insec_len is None or len(out) >= min_insec_len) and \
                (fendsec is None or fendsec(datapoint)):
                if off_sec_cnt >= min_endsec_len:
                    return out
                off_sec_cnt += 1
                in_section = False
            else:
                in_section = False
                out = []
        elif not in_section and len(out) > 0:
            if fendsec(datapoint):
                off_sec_cnt += 1
                if off_sec_cnt >= min_endsec_len:
                    return out
            else:
                out = []
                off_sec_cnt = 0
    if (fendsec is None or min_endsec_len <= 0) and \
        (min_insec_len is None or len(out) >= min_insec_len):
        return out
    return None

def get_time_diff(timepoint_1,timepoint_2):
    """returns the diffrence between two trackunit timestrings"""
    dtt1 = datetime.strptime(timepoint_1.split('.')[0],"%Y-%m-%dT%H:%M:%S")
    dtt2 = datetime.strptime(timepoint_2.split('.')[0],"%Y-%m-%dT%H:%M:%S")
    return dtt1-dtt2

def moving_avg(data,valname,alpha=0.01,in_same = False):
    """
    creates a copy of data where the valname index has a moving avg with alpha applied
    """
    if in_same:
        data2 = data
    else:
        data2 = deepcopy(data)
    def fma(_x,last):
        return alpha*_x+(1-alpha)*last
    last = float(data[0][valname])
    for datapoint in data2:
        last = fma(float(datapoint[valname]),last)
        datapoint[valname] = str(last)
    return data2

def start_end_from_tdelta(tdelta, preset_end=None):
    """
    returns start and end (as datetime) from a given timedelta
        (either type timedelta or and integer with days)
    if preset_end is set to a datetime, the end will be set to this value
    in any case it will return an end with hours, minutes, second, microseconds set to zero
    """
    if preset_end is None:
        end = datetime.now()
    else:
        end = preset_end
    end = end.replace(hour=0,minute=0,second=0,microsecond=0)
    if isinstance(tdelta,timedelta):
        start = end-tdelta
    else:
        irange = int(tdelta)
        if irange <= 0:
            return end, end
        start = end-timedelta(days=irange)
    return start, end

class SecureString:
    """This will just hide the api_key from being printed"""
    def __init__(self,_str):
        """Creates the opject with given string as data"""
        self.string = _str
    def gets(self):
        """returns the string"""
        return self.string
    def sets(self,value):
        """sets the string to value"""
        self.string = value



"""webcache module"""

import traceback
import json
import os
from os.path import join
from hashlib import md5
from pathlib import Path
import shutil
import aiohttp
import aiofiles
from aiohttp.helpers import BasicAuth

async def get_from_file(fname,dont_read=False):
    """get_from_file method"""
    try:
        if os.path.isfile(fname):
            if dont_read:
                return {}
            try:
                async with aiofiles.open(fname,encoding='utf8') as file:
                    content = await file.read()
                return json.loads(content)
            except json.JSONDecodeError:
                print("Error at TUCache get_from_file(1):\n"+str(traceback.format_exc()))
                try:
                    os.remove(fname)
                except OSError:
                    print("Error at TUCache get_from_file(3):\n"+str(traceback.format_exc()))
                return None
        else:
            return None
    except OSError:
        print("Error at TUCache get_from_file(2):\n"+str(traceback.format_exc()))
        return None

class WebCache:
    """WebCache class"""
    def __init__(self,**kwargs):
        if kwargs.get('verbose',False):
            print("WebCaches args:",kwargs)
        auth_tuple = kwargs.get("auth",None)
        self.auth = BasicAuth(auth_tuple[0].gets(),auth_tuple[1].gets()) \
            if auth_tuple is not None else None
        self.verbose = kwargs.get("verbose",False)
        self.dir = kwargs.get("webcache_dir","web-cache")
        self.dont_read_files = kwargs.get("dont_read_files",False)
        self.dont_return_data = kwargs.get("dont_return_data",False)
        self.return_only_cache_files = kwargs.get("return_only_cache_files",False)
        self.dont_cache_data = kwargs.get("dont_cache_data",False)
        Path(self.dir).mkdir(parents=True, exist_ok=True)
    def clean(self):
        """clean method"""
        try:
            shutil.rmtree(self.dir)
        except OSError:
            print("Error at TUCache clean:\n"+str(traceback.format_exc()))
        Path(self.dir).mkdir(parents=True, exist_ok=True)
    async def get_from_web(self,url: str) -> dict:
        """get_from_web method"""
        async with aiohttp.ClientSession() as session:
            async with session.request('GET', url,auth=self.auth) as response:
                response.raise_for_status()
                _j = await response.json()
                _t = await response.text()
                return _j , _t
    async def get(self,url):
        """get method"""
        if self.dont_cache_data:
            data, _ = await self.get_from_web(url)
            return data
        fname = md5(url.encode('utf-8')).hexdigest()+".json"
        if self.return_only_cache_files:
            return fname
        fname = join(self.dir,fname)
        data = await get_from_file(fname,self.dont_read_files)
        if data is None:
            data, text = await self.get_from_web(url)
            async with aiofiles.open(fname, mode='w+',encoding='utf8') as _fp:
                await _fp.write(text)
            if self.verbose:
                print(url,len(text),"W")
        else:
            if self.verbose:
                print(url,len(str(data)),"C")
        if self.dont_return_data:
            return {}
        return data



"""iterator for tucache data"""

class ReqIter:
    """iterator for web (url) requests"""
    def __init__(self,cache,requests,meta=None):
        self.cache = cache
        self.requests = requests
        self.iter_started = False
        self.meta = meta

    def __aiter__(self):
        if self.iter_started:
            raise Exception("cant start tuiter more than once")
        self.iter_started = True
        return self

    async def __anext__(self):
        try:
            return await self.cache.get_url(next(self.requests)) , self.meta
        except StopIteration as exc:
            raise StopAsyncIteration from exc

class TuIter:
    """iterator holding all internal iterators"""
    def __init__(self) -> None:
        self.iterators = []
        self.iterator_pos = 0
        self.iter_started = False

    def add(self,_iter):
        """addes an internal iterator to this iterators list"""
        self.iterators.append(_iter)

    def __aiter__(self):
        if self.iter_started:
            raise Exception("cant start tuiter more than once")
        self.iter_started = True
        return self

    async def __anext__(self):
        if len(self.iterators) == 0:
            raise StopAsyncIteration
        try:
            return await self.iterators[self.iterator_pos].__anext__()
        except StopAsyncIteration as exc:
            self.iterator_pos += 1
            if self.iterator_pos >= len(self.iterators):
                raise StopAsyncIteration from exc
            return await self.__anext__()



"""tucache module"""

from datetime import timedelta
from math import ceil
from .webcache import WebCache
from .tuiter import ReqIter, TuIter
from .helper import start_end_from_tdelta

URL_BASE = r'https://api.trackunit.com/public/'

class TuCache:
    """tucache class"""
    def __init__(self,**kwargs):
        self.cache = WebCache(**kwargs)
        self.req_period = 30
        self.tdelta_end = kwargs.get("tdelta_end",None)
    def clean(self):
        """deletes all cached data"""
        self.cache.clean()
    async def get_url(self,url):
        """takes the data from cache if possible. otherwise data is loaded from web"""
        data = await self.cache.get(URL_BASE+url)
        if self.cache.return_only_cache_files:
            return [data]
        if self.cache.dont_return_data:
            return []
        if self.cache.dont_read_files and len(data) == 0:
            return []
        return data.get('list')
    async def get_unitlist(self):
        """returns a list of vehicles"""
        return await self.get_url('Unit')
    def general_daydiff_get(self,furl,meta,previter=None):
        """returns data for timedependant requests for a given daydelta"""
        tdelta = meta["tdelta"]
        start,end = start_end_from_tdelta(tdelta,self.tdelta_end)
        meta["start"] = start
        meta["end"] = end
        return self.general_time_range_get(furl,meta,previter)
    def general_time_range_get(self,furl,meta,previter=None):
        """returns data for timedependant requests for a start and enddate"""
        start = meta["start"]
        end = meta["end"]
        days = (end-start).days
        requests = []
        for week in range(ceil(days/self.req_period)):
            wstart = start+timedelta(days=week*self.req_period)
            wend = wstart+timedelta(days=min(self.req_period,(end-wstart).days))
            requests.append(furl(\
                wstart.strftime("%Y-%m-%dT%H:%M:%S"),\
                wend.strftime("%Y-%m-%dT%H:%M:%S")))
        internal_iter = ReqIter(self,iter(requests),meta)
        if previter is None:
            previter = TuIter()
        previter.add(internal_iter)
        return previter,len(requests)

    def get_history(self,veh_id,tdelta,previter=None):
        """getHistory method"""
        meta = {}
        meta["id"] = veh_id
        meta["tdelta"] = tdelta
        return self.general_daydiff_get(lambda t1,t2: \
            'Report/UnitHistory?unitId='+veh_id+'&from='+t1+'.0000001Z&to='+t2+'.0000000Z',\
                meta,previter)
    def get_history_timedelta(self,veh_id,start,end,previter=None):
        """getHistory method"""
        meta = {}
        meta["id"] = veh_id
        meta["start"] = start
        meta["end"] = end
        return self.general_time_range_get(lambda t1,t2: \
            'Report/UnitHistory?unitId='+veh_id+'&from='+t1+'.0000001Z&to='+t2+'.0000000Z',\
                meta,previter)
    def get_candata(self,veh_id,tdelta=None,previter=None):
        """getCanData method"""
        meta = {}
        meta["id"] = veh_id
        meta["tdelta"] = tdelta
        return self.general_daydiff_get(lambda t1,t2: \
            'Report/UnitExtendedInfo?Id='+veh_id+'&from='+t1+'.0000001Z&to='+t2+'.0000000Z',\
                meta,previter)
    def get_candata_timedelta(self,veh_id,start,end,previter=None):
        """getCanData method"""
        meta = {}
        meta["id"] = veh_id
        meta["start"] = start
        meta["end"] = end
        return self.general_time_range_get(lambda t1,t2: \
            'Report/UnitExtendedInfo?Id='+veh_id+'&from='+t1+'.0000001Z&to='+t2+'.0000000Z',\
                meta,previter)
    def get_faults(self,veh_id,tdelta=None,previter=None):
        """get_faults method"""
        meta = {}
        meta["id"] = veh_id
        meta["tdelta"] = tdelta
        return self.general_daydiff_get(lambda t1,t2: \
            'Report/UnitActiveFaults?id='+veh_id+'&from='+t1+'.0000001Z&to='+t2+'.0000000Z',\
                meta,previter)
    def get_faults_timedelta(self,veh_id,start,end,previter=None):
        """get_faults method"""
        meta = {}
        meta["id"] = veh_id
        meta["start"] = start
        meta["end"] = end
        return self.general_time_range_get(lambda t1,t2: \
            'Report/UnitActiveFaults?id='+veh_id+'&from='+t1+'.0000001Z&to='+t2+'.0000000Z',\
                meta,previter)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pytrackunit",
    version="2.3.7",
    author="Marius Schlueter",
    author_email="themrslue@googlemail.com",
    description="Easy access for TrackUnit REST API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/einsteinmaster/TrackUnitPython",
    project_urls={
        "Bug Tracker": "https://github.com/einsteinmaster/TrackUnitPython/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=["pytrackunit"],
    install_requires=["aiohttp", "aiofiles", "tqdm", "python-dateutil", "aiosqlite"],
    python_requires=">=3.8",
)


"""Module for caching data in sql db"""

from copy import deepcopy
import os.path
import os
from datetime import datetime
import sqlite3
import aiosqlite
from .tucache import TuCache
from .tuiter import TuIter
from .helper import get_datetime, start_end_from_tdelta


CREATE_HISTORY_META = '''
create table historymeta(
    unit text, 
    start text, 
    end text,
    CONSTRAINT time CHECK (start < end),
    primary key(unit, start)
)
'''

CREATE_HISTORY_TABLE = '''
CREATE TABLE history(
    unit text not null, 
    time int not null, 
    event int, 
    keyId text, 
    latitude real,
    longitude real, 
    streetAddress text,
    postalCode text,
    city text,
    country text,
    heading int, 
    speed real,
    km real, 
    run1 real,
    run2 real,
    run3 real,
    run4 real,
    runOdo real,
    temperature1 real,
    temperature2 real,
    input1 int,
    input2 int,
    input3 int,
    input4 int,
    input5 int,
    input6 int,
    input7 int,
    input8 int,
    input9 int,
    input10 int,
    output1 int,
    output2 int,
    output3 int,
    output4 int,
    output5 int,
    analogInput1 real,
    analogInput2 real,
    analogInput4 real,
    Input1ChangeCounter INT,
    Input2ChangeCounter INT,
    Input3ChangeCounter INT,
    Input4ChangeCounter INT,
    batteryLevel real,
    externalPower real,
    primary key (unit,time)
)
'''

CREATE_ERROR_META_TABLE = '''
CREATE TABLE errormeta(
    unit text, 
    start int, 
    end int,
    CONSTRAINT time CHECK (start < end),
    primary key(unit, start)
)
'''

CREATE_ERROR_DATA_TABLE = '''
CREATE TABLE error(
    unit text not null, 
    time int not null, 
    spn int not null,
    fmi int not null,
    occurenceCount int,
    name text,
    description text,
    primary key(unit,time,spn,fmi)
)
'''

CREATE_CANDATA_META_TABLE = '''
CREATE TABLE candatameta(
    unit text, 
    start int, 
    end int,
    CONSTRAINT time CHECK (start < end),
    primary key(unit, start)
)
'''

CREATE_CANDATA_DATA_TABLE = '''
CREATE TABLE candata(
    unit text not null, 
    time int not null, 
    variableId int not null,
    name text not null,
    value text not null,
    uoM text,
    primary key(unit,time,variableId)
)
'''

def create_tables(db_path):
    """Creates the necessary tables in an sqlite database which is located at db_path"""
    _db = sqlite3.connect(db_path)
    cur = _db.cursor()
    cur.execute(CREATE_HISTORY_TABLE)
    cur.execute(CREATE_HISTORY_META)
    cur.execute(CREATE_ERROR_META_TABLE)
    cur.execute(CREATE_ERROR_DATA_TABLE)
    cur.execute(CREATE_CANDATA_META_TABLE)
    cur.execute(CREATE_CANDATA_DATA_TABLE)
    _db.commit()
    _db.close()

def candata_item_to_sql_item(_x,meta):
    """
    returns the candata as a tuple and converts the time to unix timestamp (milliseconds)
    """
    _id = meta['id']
    _time = int(get_datetime(_x['time']).timestamp()*1000)
    _variableid = _x['variableId']
    _name = _x['name']
    _value = _x['value']
    if 'uoM' in _x:
        _uom = _x['uoM']
    else:
        _uom = None
    return (_id,_time,_variableid,_name,_value,_uom)

def sql_item_to_candata_item(obj):
    """
    the operation candata_item_to_sql_item reversed
    """
    _x = {}
    _x['id'] = obj[0]
    _x['time'] = datetime.fromtimestamp(obj[1]/1000.0).isoformat()
    _x['variableId'] = obj[2]
    _x['name'] = obj[3]
    _x['value'] = obj[4]
    _x['uoM'] = obj[5]
    return _x

def error_item_to_sql_item(_x,meta):
    """
    returns the error as a tuple and converts the time to unix timestamp (milliseconds)
    """
    _id = meta['id']
    #print(x['time'])
    _time = int(get_datetime(_x['time']).timestamp()*1000)
    #print(_time,x['time'],get_datetime(x['time']))
    _spn = _x['spn']
    _fmi = _x['fmi']
    _oc = _x['occurrenceCount']
    if 'name' in _x:
        _name = _x['name']
    else:
        _name = None
    if 'description' in _x:
        _desc = _x['description']
    else:
        _desc = None
    return (_id,_time,_spn,_fmi,_oc,_name,_desc)

def sql_item_to_error_item(obj):
    """
    the operation error_item_to_sql_item reversed
    """
    _x = {}
    _x['id'] = obj[0]
    _x['time'] = datetime.fromtimestamp(obj[1]/1000.0).isoformat()
    _x['spn'] = obj[2]
    _x['fmi'] = obj[3]
    _x['occurrenceCount'] = obj[4]
    _x['name'] = obj[5]
    _x['description'] = obj[6]
    return _x

# pylint: disable=invalid-name, too-many-locals
def history_item_to_sql_item(_x,meta):
    """
    returns the history as a tuple and converts the time to unix timestamp (milliseconds)
    """
    _time = int(get_datetime(_x['time']).timestamp()*1000)
    _unit = meta['id']
    _event = _x['event'] if 'event' in _x else None
    _keyId = _x['keyId'] if 'keyId' in _x else None
    _latitude = _x['latitude'] if 'latitude' in _x else None
    _longitude = _x['longitude'] if 'longitude' in _x else None
    _streetAddress = _x['streetAddress'] if 'streetAddress' in _x else None
    _postalCode = _x['postalCode'] if 'postalCode' in _x else None
    _city = _x['city'] if 'city' in _x else None
    _country = _x['country'] if 'country' in _x else None
    _heading = _x['heading'] if 'heading' in _x else None
    _speed = _x['speed'] if 'speed' in _x else None
    _km = _x['km'] if 'km' in _x else None
    _run1 = _x['run1'] if 'run1' in _x else None
    _run2 = _x['run2'] if 'run2' in _x else None
    _run3 = _x['run3'] if 'run3' in _x else None
    _run4 = _x['run4'] if 'run4' in _x else None
    _runOdo = _x['runOdo'] if 'runOdo' in _x else None
    _temperature1 = _x['temperature1'] if 'temperature1' in _x else None
    _temperature2 = _x['temperature2'] if 'temperature2' in _x else None
    _input1 = _x['input1'] if 'input1' in _x else None
    _input2 = _x['input2'] if 'input2' in _x else None
    _input3 = _x['input3'] if 'input3' in _x else None
    _input4 = _x['input4'] if 'input4' in _x else None
    _input5 = _x['input5'] if 'input5' in _x else None
    _input6 = _x['input6'] if 'input6' in _x else None
    _input7 = _x['input7'] if 'input7' in _x else None
    _input8 = _x['input8'] if 'input8' in _x else None
    _input9 = _x['input9'] if 'input9' in _x else None
    _input10 = _x['input10']  if 'input10' in _x else None
    _output1 = _x['output1'] if 'output1' in _x else None
    _output2 = _x['output2'] if 'output2' in _x else None
    _output3 = _x['output3'] if 'output3' in _x else None
    _output4 = _x['output4'] if 'output4' in _x else None
    _output5 = _x['output5'] if 'output5' in _x else None
    _analogInput1 = _x['analogInput1'] if 'analogInput1' in _x else None
    _analogInput2 = _x['analogInput2'] if 'analogInput2' in _x else None
    _analogInput4 = _x['analogInput4'] if 'analogInput4' in _x else None
    _Input1ChangeCounter = _x['Input1ChangeCounter'] if 'Input1ChangeCounter' in _x else None
    _Input2ChangeCounter = _x['Input2ChangeCounter'] if 'Input2ChangeCounter' in _x else None
    _Input3ChangeCounter = _x['Input3ChangeCounter'] if 'Input3ChangeCounter' in _x else None
    _Input4ChangeCounter = _x['Input4ChangeCounter'] if 'Input4ChangeCounter' in _x else None
    _batteryLevel = _x['batteryLevel'] if 'batteryLevel' in _x else None
    _externalPower = _x['externalPower'] if 'externalPower' in _x else None

    return (_unit, _time, _event, _keyId, _latitude, _longitude,
        _streetAddress, _postalCode, _city, _country, _heading,
        _speed, _km, _run1, _run2, _run3, _run4, _runOdo, _temperature1,
        _temperature2, _input1, _input2, _input3, _input4, _input5,
        _input6, _input7, _input8, _input9, _input10, _output1, _output2,
        _output3, _output4, _output5, _analogInput1, _analogInput2,
        _analogInput4, _Input1ChangeCounter, _Input2ChangeCounter,
        _Input3ChangeCounter, _Input4ChangeCounter,
        _batteryLevel, _externalPower)
# pylinte: enable=invalid-name, too-many-locals

def sql_item_to_history_item(obj):
    """
    the operation history_item_to_sql_item reversed
    """
    _x = {}
    _x['unit'] = obj[0]
    _x['time'] = datetime.fromtimestamp(obj[1]/1000.0).isoformat()
    _x['event'] = obj[2]
    _x['keyId'] = obj[3]
    _x['latitude'] = obj[4]
    _x['longitude'] = obj[5]
    _x['streetAddress'] = obj[6]
    _x['postalCode'] = obj[7]
    _x['city'] = obj[8]
    _x['country'] = obj[9]
    _x['heading'] = obj[10]
    _x['speed'] = obj[11]
    _x['km'] = obj[12]
    _x['run1'] = obj[13]
    _x['run2'] = obj[14]
    _x['run3'] = obj[15]
    _x['run4'] = obj[16]
    _x['runOdo'] = obj[17]
    _x['temperature1'] = obj[18]
    _x['temperature2'] = obj[19]
    _x['input1'] = obj[20]
    _x['input2'] = obj[21]
    _x['input3'] = obj[22]
    _x['input4'] = obj[23]
    _x['input5'] = obj[24]
    _x['input6'] = obj[25]
    _x['input7'] = obj[26]
    _x['input8'] = obj[27]
    _x['input9'] = obj[28]
    _x['input10'] = obj[29]
    _x['output1'] = obj[30]
    _x['output2'] = obj[31]
    _x['output3'] = obj[32]
    _x['output4'] = obj[33]
    _x['output5'] = obj[34]
    _x['analogInput1'] = obj[35]
    _x['analogInput2'] = obj[36]
    _x['analogInput4'] = obj[37]
    _x['Input1ChangeCounter'] = obj[38]
    _x['Input2ChangeCounter'] = obj[39]
    _x['Input3ChangeCounter'] = obj[40]
    _x['Input4ChangeCounter'] = obj[41]
    _x['batteryLevel'] = obj[42]
    _x['externalPower'] = obj[43]
    return _x

async def SqlInsertIter(dbpath, table, sqliter, meta, verbose=False):
    """Generator which inserts data from an upstream iterator into the database"""
    assert dbpath is not None
    assert dbpath != ""

    async with aiosqlite.connect(dbpath) as _db:

        if verbose:
            print("SqlInsertIter __init__ has sqliter:",sqliter)
        if table == "error":
            insert_sql = "INSERT INTO error VALUES (?,?,?,?,?,?,?)"
            fconv = error_item_to_sql_item
        elif table == "history":
            insert_sql = """
            INSERT INTO history VALUES 
            (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)
            """
            fconv = history_item_to_sql_item
        elif table == "candata":
            insert_sql = """
            INSERT INTO candata VALUES 
            (?,?,?,?,?,?)
            """
            fconv = candata_item_to_sql_item
        else:
            raise Exception("Not yet implemented")

        await _db.execute(f"INSERT INTO {table}meta VALUES (?,?,?)",\
            (meta['id'],meta['start_ts'],meta['end_ts']))

        async for data,meta in sqliter:
            data = list(data)
            sqldata = map(lambda x: fconv(x[0],x[1]),zip(data,[meta]))

            try:
                await _db.executemany(insert_sql,sqldata)
            except sqlite3.IntegrityError as exc1:
                if verbose:
                    print("Integrety error with",meta)
                    print("Try to find double entry (insert side)")
                    for in_item in sqldata:
                        try:
                            await _db.execute(insert_sql,in_item)
                        except sqlite3.IntegrityError:
                            print("Item throwing Integrity error",in_item)
                await _db.rollback()
                raise sqlite3.IntegrityError from exc1

            yield data, meta

        if verbose:
            await _db.commit()
            print("Committed")



async def SqlReturnIter(db_path, table, meta):
    """Generator return given data from database"""
    command = f"select * from {table} where unit = ? and time >= ? and time <= ? order by time"
    if table == "error":
        fconv = sql_item_to_error_item
    elif table == "history":
        fconv = sql_item_to_history_item
    elif table == "candata":
        fconv = sql_item_to_candata_item
    else:
        raise Exception("Table not implemented yet")
    async with aiosqlite.connect(db_path) as _db:
        async with _db.execute(command,(meta['id'],meta['start_ts'],meta['end_ts'])) as _cur:
            async for x in _cur:
                yield [fconv(x)], meta

class SqlCache:
    """Sql cache can cache trackunit data in Sqlite DB"""
    def __init__(self,**kwargs):
        self.verbose = kwargs.get('verbose',False)
        self.db_path = kwargs.get('db_path','sqlcache.sqlite')
        self.tdelta_end = kwargs.get('tdelta_end',None)
        cache1_kwargs = deepcopy(kwargs)
        cache1_kwargs['dont_cache_data'] = kwargs.get('sql_cache1_dont_cache_data',True)
        self.cache1 = kwargs.get('sql_cache1',TuCache(**cache1_kwargs))
        self.cache2 = kwargs.get('sql_cache2',TuCache(**kwargs))
        self.db_connection = None
        self.connect()

    def connect(self):
        """Connects the SqlCache and sets up db_connection"""
        if self.db_connection is not None:
            raise Exception("Cant connect twice")
        no_db_present = not os.path.isfile(self.db_path)
        self.db_connection = sqlite3.connect(self.db_path)
        if no_db_present:
            create_tables(self.db_path)

    def close(self):
        """closes database connection"""
        if self.db_connection is None:
            return
        self.db_connection.close()
        self.db_connection = None

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.close()

    def reset(self,remove_data=False):
        """removes database file"""
        self.close()
        if remove_data:
            os.remove(self.db_path)
        self.connect()

    async def get_unitlist(self):
        """returns a list of vehicles"""
        return await self.cache2.get_unitlist()

    # pylint: disable=too-many-arguments
    def get_general_upstream(self, table, veh_id, start_ts, end_ts, previter=None):
        """gets errors from upstream cache"""

        start = datetime.fromtimestamp(start_ts/1000.0)
        end = datetime.fromtimestamp(end_ts/1000.0)

        if table == "error":
            int_data_iter, _len = self.cache1.get_faults_timedelta(veh_id,start,end,None)
        elif table == "history":
            int_data_iter, _len = self.cache1.get_history_timedelta(veh_id,start,end,None)
        elif table == "candata":
            int_data_iter, _len = self.cache1.get_candata_timedelta(veh_id,start,end,None)
        else:
            raise Exception("Not Implemented")

        meta = {}
        meta["id"] = veh_id
        meta["start_ts"] = start_ts
        meta["end_ts"] = end_ts
        meta["start"] = start
        meta["end"] = end

        wrap_iter = SqlInsertIter(self.db_path,table,int_data_iter,meta,self.verbose)

        if previter is None:
            previter = TuIter()

        previter.add(wrap_iter)

        return previter, _len
    def get_general_sql(self, table, veh_id, start_ts, end_ts, previter=None):
        """gets data of this period from db whether or not it was actually stored there"""

        cur = self.db_connection.execute(f"""
            select count(*) from {table} where
            unit = ? and time >= ? and time <= ?
            order by time
            """,(veh_id,start_ts,end_ts))
        cnt = cur.fetchone()[0]

        if self.verbose:
            print("found",cnt,"in sql")

        meta = {}
        meta["id"] = veh_id
        meta["start_ts"] = start_ts
        meta["end_ts"] = end_ts

        if previter is None:
            previter = TuIter()

        if cnt > 0:
            previter.add(SqlReturnIter(self.db_path,table,meta))
        elif self.verbose:
            print("could not find any item in block ", start_ts, end_ts,"for unit",veh_id)

        return previter, cnt
    def get_general_unixts(self, table, veh_id, start_ts, end_ts, previter=None):
        """returns error in between the given datetime objects"""

        # Query meta database to check whether there the data is in database
        cur = self.db_connection.execute(f"""
            select * from {table}meta where unit = ? and (
            (start <= ? and end > ?) or 
            (start < ? and end >= ?) or
            (start >= ? and end <= ?)
            ) order by start
            """,(veh_id,start_ts,start_ts,end_ts,end_ts,start_ts,end_ts))
        first = cur.fetchone()
        if first is not None:
            me_vehid, me_start, me_end = first[0], first[1], first[2]
        else:
            if self.verbose:
                print("Stop iteration. Didnt find",\
                    veh_id,start_ts,end_ts,\
                    "Get errors from upstream now")
            return self.get_general_upstream(table,veh_id,start_ts,end_ts,previter)

        me_start = int(float(me_start))
        me_end = int(float(me_end))

        if self.verbose:
            print("Found block",me_vehid,me_start,me_end)

        # Depending on the start and end of the next block, apply divide and conquer
        # by recursive calls of this function.
        if me_start <= start_ts:
            if me_end < end_ts:
                previter, cnt1 = self.get_general_unixts(table,veh_id,start_ts,me_end,previter)
                previter, cnt2 = self.get_general_unixts(table,veh_id,me_end+1,end_ts,previter)
                return previter,(cnt1+cnt2)
            return self.get_general_sql(table,veh_id,start_ts,end_ts,previter)
        if me_end >= end_ts:
            previter, cnt1 = self.get_general_unixts(table,veh_id,start_ts,me_start-1,previter)
            previter, cnt2 = self.get_general_unixts(table,veh_id,me_start,end_ts,previter)
            return previter,(cnt1+cnt2)
        previter, cnt1 = self.get_general_unixts(table,veh_id,start_ts,me_start-1,previter)
        previter, cnt2 = self.get_general_unixts(table,veh_id,me_start,me_end,previter)
        previter, cnt3 = self.get_general_unixts(table,veh_id,me_end+1,end_ts,previter)
        return previter,(cnt1+cnt2+cnt3)
    # pylint: enable=too-many-arguments
    def get_faults_timedelta(self,veh_id,start,end,previter=None):
        """returns error in between the given datetime objects"""
        start_ts = int(start.timestamp()*1000)
        end_ts = int(end.timestamp()*1000)
        return self.get_general_unixts("error",veh_id, start_ts, end_ts, previter)
    def get_faults(self,veh_id,tdelta=None,previter=None):
        """get_faults method"""
        start, end = start_end_from_tdelta(tdelta,self.tdelta_end)
        return self.get_faults_timedelta(veh_id,start,end,previter)
    def get_history_timedelta(self,veh_id,start,end,previter=None):
        """returns error in between the given datetime objects"""
        start_ts = int(start.timestamp()*1000)
        end_ts = int(end.timestamp()*1000)
        return self.get_general_unixts("history",veh_id, start_ts, end_ts, previter)
    def get_history(self,veh_id,tdelta=None,previter=None):
        """get_faults method"""
        start, end = start_end_from_tdelta(tdelta,self.tdelta_end)
        return self.get_history_timedelta(veh_id,start,end,previter)
    def get_candata_timedelta(self,veh_id,start,end,previter=None):
        """returns error in between the given datetime objects"""
        start_ts = int(start.timestamp()*1000)
        end_ts = int(end.timestamp()*1000)
        return self.get_general_unixts("candata",veh_id, start_ts, end_ts, previter)
    def get_candata(self,veh_id,tdelta=None,previter=None):
        """get_faults method"""
        start, end = start_end_from_tdelta(tdelta,self.tdelta_end)
        return self.get_candata_timedelta(veh_id,start,end,previter)



"""module TrackUnit"""

import json
import os.path
import asyncio
import tqdm
from .tucache import TuCache
from .sqlcache import SqlCache
from .helper import SecureString

def get_multi_general(func,idlist,tdelta,f_process=None,progress_bar=True):
    """
    returns the data of a list of vehicles with the ids provided in idlist.
    f_process can be specified to process slices of data. f_process returns a list
    """
    if f_process is None:
        f_process = lambda x, meta: x

    async def get_data_async(globit,globlen):
        _cor = []
        if progress_bar:
            pbar = tqdm.tqdm(total=globlen)
        async for _f, meta in globit:
            _cor += f_process(_f, meta)
            if progress_bar:
                pbar.update()
        return _cor

    globlen = 0
    last = None
    for _id in idlist:
        last,_l = func(_id,tdelta,last)
        globlen += _l

    data = asyncio.run(get_data_async(last,globlen))

    return data

class TrackUnit:
    """TrackUnit class"""
    def __init__(self,**kwargs):
        config_filename = kwargs.get("config_path","config.json")
        config = {}
        if os.path.isfile(config_filename):
            with open(config_filename,encoding="utf8") as file:
                config = json.load(file)
        for k in config:
            if k not in kwargs:
                kwargs[k] = config[k]
        if 'auth' not in kwargs:
            if kwargs.get('api_key',None) is None:
                if 'apikey_path' not in kwargs:
                    kwargs['apikey_path'] = 'api.key'
                with open(kwargs["apikey_path"],encoding="utf8") as file_apikey:
                    api_key = SecureString(file_apikey.readline())
            else:
                api_key = kwargs['api_key']
            if len(api_key.gets()) != 32:
                raise Exception("Invalid API-key length")
            kwargs['auth'] = (SecureString('API'),api_key)
        if kwargs.get('tu_use_sqlcache',False):
            self.cache = SqlCache(**kwargs)
        else:
            self.cache = TuCache(**kwargs)
    async def _a_get_unitlist(self,_type=None,sort_by_hours=True):
        """unitList method"""
        data = await self.cache.get_unitlist()
        if _type is not None:
            data = list(filter(lambda x: " " in x['name'] and _type in x['name'],data))
        if sort_by_hours:
            if not isinstance(data,list):
                data = list(data)
            data.sort(key=lambda x: (x['run1'] if 'run1' in x else 0),reverse=True)
        return data

    def get_unitlist(self,_type=None,sort_by_hours=True):
        """unitList method"""
        return asyncio.run(self._a_get_unitlist(_type,sort_by_hours))

    async def _a_get_history(self,veh_id,tdelta):
        """async getHistory method"""
        data = []
        _it, _ = self.cache.get_history(veh_id,tdelta)
        async for _d,_ in _it:
            data += _d
        return data

    async def _a_get_candata(self,veh_id,tdelta=None):
        """async getCanData method"""
        data = []
        _it, _ = self.cache.get_candata(veh_id,tdelta)
        async for _d,_ in _it:
            data += _d
        return data

    async def _a_get_faults(self,veh_id,tdelta=None):
        """async get_faults method"""
        data = []
        _it, _ = self.cache.get_faults(veh_id,tdelta)
        async for _d,_ in _it:
            data += _d
        return data

    def get_history(self,veh_id,tdelta):
        """getHistory method"""
        return asyncio.run(self._a_get_history(veh_id,tdelta))

    def get_candata(self,veh_id,tdelta=None):
        """getCanData method"""
        return asyncio.run(self._a_get_candata(veh_id,tdelta))

    def get_faults(self,veh_id,tdelta=None):
        """get_faults method"""
        return asyncio.run(self._a_get_faults(veh_id,tdelta))

    def get_multi_history(self,idlist,tdelta,f_process=None,progress_bar=True):
        """
        returns the data of a list of vehicles with the ids provided in idlist.
        f_process can be specified to process slices of data. f_process returns a list
        """
        return get_multi_general(
            self.cache.get_history,idlist,tdelta,f_process,progress_bar)

    def get_multi_candata(self,idlist,tdelta,f_process=None,progress_bar=True):
        """
        returns the data of a list of vehicles with the ids provided in idlist.
        f_process can be specified to process slices of data. f_process returns a list
        """
        return get_multi_general(
            self.cache.get_candata,idlist,tdelta,f_process,progress_bar)

    def get_multi_faults(self,idlist,tdelta,f_process=None,progress_bar=True):
        """
        returns the data of a list of vehicles with the ids provided in idlist.
        f_process can be specified to process slices of data. f_process returns a list
        """
        return get_multi_general(
            self.cache.get_faults,idlist,tdelta,f_process,progress_bar)
